#!/usr/bin/env python
"""
The client server-side-backups app is tool to manage your router's backups and thus it is a proxy to call remote API.

    Copyright (C) 2017 CZ.NIC, z.s.p.o. (http://www.nic.cz/)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

It provides these functions :

 * authentication against server of server-side-backups app
 * encrypt/decrypt backup file
 * call remote API to list, retrieve, delete, create router backups

It always returns exit code :
 0 - everything should be ok
 -1 .. -10 -> there is an error in current program run

When exit code is not zero than every time will appear JSON in stdout in the format:

 {"detail": "lorem ipsum, lorem ipsum...", "code": -1, "type": "ArgumentError"}

It always return additionaly data in JSON.
"""

import os
import json
import getpass
import re
from sys import stdout, stdin
from argparse import ArgumentParser
from subprocess import PIPE, Popen, check_output, CalledProcessError


ERROR_TYPES = ('ClientException', 'ArgumentTypeError', 'ArgumentError', 'CalledProcessError', 'UnknownException')


###
# error code meanings
#   - Remote API error
ERR_CODE_API_CALL = -1
#   - Script arguments error
ERR_CODE_ARGS = -2
#   - Arguments validation error
ERR_CODE_VALIDATE = -3
#   - Registration code error
ERR_CODE_REG_CODE = -4
#   - Encrypting error
ERR_CODE_ENCRYPT = -5
#   - Decrypting error
ERR_CODE_DECRYPT = -6
#   - Password from prompt mismatched
ERR_PASSWD_MISMATCHED = -7
#   - Unknown error
ERR_CODE_UNKNOWN = -10


class SSBackupsException(Exception):
    """
    General client exception for this module. It has 3 attributtes.
     * detail -> verbal explanation of the exception,
     * code -> a number expressing type of exception and a exit code for this script.
     * type -> real type of the exception -> builtin exception types are mapped to SSBackupsException,
               it can be any value of ERROR_TYPES variable"""

    def __init__(self, *args):
        super(SSBackupsException, self).__init__(*args)
        self.detail = args[0]
        self.code = ERR_CODE_API_CALL if len(args) == 1 else args[1]
        self.type = ERROR_TYPES[0] if len(args) <= 2 else args[2]


def type_filepath(value):
    """
    Make sure that value is filepath or SSBackupsException is raised.
    :param value: a file path
    :return: a file path
    """
    if os.path.isfile(value):
        return value
    raise SSBackupsException("Given path is not the existing path to any file!", ERR_CODE_ARGS, ERROR_TYPES[1])


def type_target_path(value):
    """
    Make sure that value is file path and directory of that file exists or SSBackupsException is raised.
    :param value: a directory path
    :return: a directory path
    """
    if not value.split(os.path.sep)[-1]:
        raise SSBackupsException("Given file path needn't slash at the end of the path!", ERR_CODE_ARGS, ERROR_TYPES[1])
    if not os.path.isdir(os.path.dirname(value)):
        raise SSBackupsException("Given file path has no existing directory!", ERR_CODE_ARGS, ERROR_TYPES[1])
    return value


def type_url(value):
    """
    Make sure that value is URL in form or meaning of scheme://hostname:port
    :param value:  Remote URL
    :return: Remote URL
    """
    error = False
    url_parts = value.split(':')
    if url_parts[0] not in ('http', 'https'):
        # URL scheme is not valid
        error = True
    elif 1 < len(url_parts) < 4:
        if not url_parts[1].startswith('//'):
            # before hostname must be double slash
            error = True
        if len(url_parts) == 3:
            if not re.match(ur"[0-9]{2,5}", url_parts[2]):
                # port is not a number
                error = True
    else:
        # count of parts of URL is not valid
        error = True
    if error:
        raise SSBackupsException("Bad value of URL!", ERR_CODE_ARGS, ERROR_TYPES[1])
    return value


# It maps method and URL path to required action
API_MAPPING = {
    'list': (
        'GET',
        '/backups/',
    ),
    'create': (
        'POST',
        '/backups/',
    ),
    'delete': (
        'DELETE',
        '/backups/{backup_id}/',
    ),
    'retrieve': (
        'GET',
        '/backups/{backup_id}/',
    ),
    'ondemand': (
        'PUT',
        '/backups/{backup_id}/on-demand/'
    )
}


def get_password(tty_recheck=False):
    """
    It gets password from stdin.

    :param tty_recheck: force reenter password when the program is used on tty
    :type tty_recheck: bool

    :return: password
    :rtype: str
    :raises SSBackupsException: when passwords mismatch
    """
    if stdin.isatty():
        # A password proper prompt
        passwd = getpass.getpass('Password for backup: ')
        if tty_recheck and (passwd != getpass.getpass('Password for backup (again): ')):
            raise SSBackupsException(
                "Passwords inserted from prompt mismatched!",
                ERR_PASSWD_MISMATCHED,
                ERROR_TYPES[0],
            )
    else:
        # Read stdin directly when the program is triggered using pipe
        passwd = stdin.readline().rstrip()

    return passwd


def remove_encrypted_backup(backup):
    """
    It removes file of encrypted backup from FS
    :param backup: a file path where backup is or will be stored
    :return: None
    """
    os.remove(backup + '.gpg')


def encrypt_backup(backup, password):
    """
    It encryptes backup with password
    :param backup: a file path to backup
    :param password: passphrase
    :return: None
    """
    ###
    # Explanation of used gpg command arguments
    #
    # --yes                 - rewrite target file for case when it has already existed
    #
    # --batch               - Use  batch  mode.  Never ask, do not allow interactive commands.
    #                         Hide the prompt : "Reading passphrase from file descriptor 0..."
    #
    # --passphrase-fd 0     - Read the passphrase from file descriptor 0. Only the first line will be read.
    #                         The passphrase will be read from STDIN.
    #
    # --cipher-algo=AES256  - Use AES256 as cipher algorithm (default is AES128).
    #
    # -o                    - Write output to file.
    #
    # -c                    - Encrypt with a symmetric cipher using a passphrase.
    ###
    cmd = [
        "gpg", "--batch", "--yes", "--passphrase-fd", "0", "--cipher-algo=AES256",
        "-o", "{backup_path}.gpg".format(backup_path=backup),
        "-c", backup
    ]
    try:
        ###
        # stdin=PIPE -> use to pass password to gpg command
        # stderr=PIPE -> to catch error
        ###
        process = Popen(cmd, stdin=PIPE, stderr=PIPE)
        error = process.communicate(password)[1]
        if process.returncode != 0:
            raise SSBackupsException(
                "Command << {cmd} >> failed having this on stderr << {error} >>.".format(
                    cmd=cmd,
                    error=error
                ), ERR_CODE_ENCRYPT, ERROR_TYPES[4])
    except CalledProcessError as exc:
        raise SSBackupsException(
            "Command << {cmd} >> ended with << return code = {return_code} >> and with output << {output} >>.".format(
                cmd=exc.cmd,
                return_code=exc.returncode,
                output=str(exc.output)
            ), ERR_CODE_ENCRYPT, ERROR_TYPES[3])


def decrypt_backup(target, backup, password):
    """
    It decryptes backup with password
    :param target: where to store backup after downloading and decrypting
    :param backup: content of backup
    :param password: passphrase
    :return: None
    """
    tmp_file = open(target + '.gpg', 'wb')
    tmp_file.write(backup)
    tmp_file.close()
    ###
    # Explanation of used gpg command arguments
    #
    # --yes                 - rewrite target file for case when it has already existed
    #
    # --batch               - Use  batch  mode.  Never ask, do not allow interactive commands.
    #                         Hide the prompt : "Reading passphrase from file descriptor 0..."
    #
    # --passphrase-fd 0     - Read the passphrase from file descriptor 0. Only the first line will be read.
    #                         The passphrase will be read from STDIN.
    #
    # -o                    - Write output to file.
    #
    # -d                    - Decrypt with a symmetric cipher using a passphrase.
    ###
    cmd = [
        "gpg", "--batch", "--yes", "--passphrase-fd", "0",
        "-o", target,
        "-d", "{target_path}.gpg".format(target_path=target)
    ]
    try:
        ###
        # stdin=PIPE -> use to pass password to gpg command
        # stderr=PIPE -> to catch error
        ###
        process = Popen(cmd, stdin=PIPE, stderr=PIPE)
        error = process.communicate(password)[1]
        if process.returncode != 0:
            raise SSBackupsException(
                "Command << {cmd} >> failed having this on stderr << {error} >>.".format(
                    cmd=cmd,
                    error=error
                ), ERR_CODE_DECRYPT, ERROR_TYPES[4])
    except CalledProcessError as exc:
        raise SSBackupsException(
            "Command << {cmd} >> ended with << return code = {return_code} >> and with output << {output} >>.".format(
                cmd=exc.cmd,
                return_code=exc.returncode,
                output=str(exc.output)
            ), ERR_CODE_DECRYPT, ERROR_TYPES[3])


def get_registration_code(reg_code_path):
    """
    It returns registration code of router
    :param reg_code_path: a path to file where registration code is stored
    :return: string registration code
    """
    try:
        return open(reg_code_path, 'r').read()
    except IOError as exc:
        raise SSBackupsException(
            "Attempt to get registration code failed with message << {message} >>".format(
                message=exc.message
            ), ERR_CODE_REG_CODE, ERROR_TYPES[3])


# pylint: disable=too-many-arguments
def call_rest_api(reg_code, url, action, backup_id=None, backup=None, connection_timeout=10):
    """
    Wrapper to call remote API
    :param reg_code: a router registration code
    :param url: URL of REST API server-side-backups
    :param action: one of 'list', 'create', 'retrieve', 'delete', 'on-demand'
    :param backup_id: id of backup stored in remote database
    :param backup: a file path of backup where it is stored on local FS
    :param connection_timeout: a network layer timeout in seconds
    :return: response (in most cases JSON = action must be in ('list', 'create', 'ondemand'))
    :raises: SSBackupsException: when anything went wrong
    """
    # prepare data to use with curl
    # method = HTTP method, template_path = template of REST API path
    method, template_path = API_MAPPING[action]
    # path = real REST API path
    path = template_path.format(backup_id=backup_id)
    # form data = definition of payload to send it to server side via REST API
    form_data = ["-F", "payload=@%s.gpg" % backup] if backup else []
    # rest_api_url = real REST API URL
    rest_api_url = "{url}{path}".format(url=url, path=path)

    ###
    # build command curl to able to call SSBackups REST API
    # --fail triggers exit code of curl to appear when remote HTTP source returns 4xx or 5xx HTTP status codes.
    ###
    cmd = [
        "/usr/bin/curl", "--fail", "-X", method, "-m", str(connection_timeout),
        "-H", "Accept:application/json", "-H", "Authorization:Token %s" % reg_code
    ] + form_data + [rest_api_url]
    try:
        ###
        # stderr=open('/dev/null','w') is here to hide progress meter
        ###
        return check_output(cmd, stderr=open('/dev/null', 'w'))
    except CalledProcessError as exc:
        raise SSBackupsException(
            "Command << {cmd} >> ended with << return code = {return_code} >> and with output << {output} >>.".format(
                cmd=exc.cmd,
                return_code=exc.returncode,
                output=str(exc.output)
            ), ERR_CODE_API_CALL, ERROR_TYPES[0])


def backup_create(reg_code, url, backup, password, connection_timeout=10):
    """
    Wrapper to action create and its call_rest_api
    :param reg_code: router registration code(16 bytes length)
    :param url: url of remote REST API
    :param backup: a file path to backup
    :param password: passphrase for gpg to encrypt backup with AES256
    :param connection_timeout: network layer timeout in seconds
    :return: string: JSON response {'id': n}, where n is positive integer
    :raises: SSBackupsException: when anything went wrong
    """

    encrypt_backup(backup, password or get_password(tty_recheck=True))

    res = call_rest_api(
        reg_code,
        url,
        'create',
        backup=backup,
        connection_timeout=connection_timeout
    )

    remove_encrypted_backup(backup)

    return res


def backup_retrieve(reg_code, url, backup_id, target, password, connection_timeout=10):
    """
    Wrapper to action retrieve and its call_rest_api
    :param reg_code: router registration code(16 bytes length)
    :param url: url of remote REST API
    :param backup_id: id of backup stored in remote database
    :param target: a file path where backup should be stored on local FS
    :param password: passphrase for gpg to decrypt backup
    :param connection_timeout: network layer timeout in seconds
    :return: None
    :raises: SSBackupsException: when anything went wrong
    """
    res = call_rest_api(
        reg_code,
        url,
        'retrieve',
        backup_id=backup_id,
        connection_timeout=connection_timeout
    )

    decrypt_backup(target, res, password or get_password())

    remove_encrypted_backup(target)


if __name__ == '__main__':
    # exit code when it is sunny days
    exit_code = 0

    try:
        parser = ArgumentParser(description=("It manages backups of Turris router. It allows to list, create, delete, "
                                             "retrieve and mark stable router's backups. Behind the every each action "
                                             "is remote API call. Password to encrypt or decrypt backup can be passed "
                                             "via -w option or via prompt while command runs."))
        parser.add_argument(
            '-t', '--connection-timeout',
            dest='connection_timeout',
            default=10,
            type=int,
            metavar='SECONDS',
            help="Timeout for http requests"
        )
        parser.add_argument(
            '-u', '--url',
            dest='url',
            default='https://rb.turris.cz',
            type=type_url,
            metavar='URL',
            help="URL of remote server-side-backups API - scheme://hostname:port"
        )
        parser.add_argument(
            '-r', '--reg-code-path',
            dest='reg_code_path',
            default='/usr/share/server-uplink/registration_code',
            type=type_filepath,
            metavar='FILE',
            help="File path to registration code"
        )
        parser.add_argument(
            '-w', '--passwd',
            dest='passwd',
            default=None,
            type=str,
            metavar='PASSWORD',
            help="Password to encrypt a backup"
        )

        # subparsers to dinstinguish between actions
        subparsers = parser.add_subparsers(
            description="Managing router backups on the server side",
            help="Possible actions to do"
        )
        # action --list does not need additional arguments
        parser_list = subparsers.add_parser('list', help="list all of the backups on the server-side")
        parser_list.set_defaults(action='list')

        # action --create requires backup argument
        parser_create = subparsers.add_parser('create', help="encrypt and upload backup to server side")
        parser_create.set_defaults(action='create')
        parser_create.add_argument(
            'backup',
            type=type_filepath,
            metavar='FILE',
            help="Path to a backup file"
        )

        # action --delete requires backup_id argument
        parser_delete = subparsers.add_parser('delete', help="delete backup forever on server side")
        parser_delete.set_defaults(action='delete')
        parser_delete.add_argument(
            'backup_id',
            type=int,
            metavar='ID',
            help="Id of a backup known to server-side-backups server",
        )

        # action --retrieve requires backup_id and target arguments
        parser_retrieve = subparsers.add_parser('retrieve', help="download backup and decrypt from server side")
        parser_retrieve.set_defaults(action='retrieve')
        parser_retrieve.add_argument(
            'backup_id',
            type=int,
            metavar='ID',
            help="Id of a backup known to server-side-backups server",
        )
        parser_retrieve.add_argument(
            'target',
            type=type_target_path,
            metavar='FILE',
            help="Path to a file where backup will be saved"
        )

        # action --ondemand requires backup_id
        parser_ondemand = subparsers.add_parser('ondemand', help="mark backup as stable on server side")
        parser_ondemand.set_defaults(action='ondemand')
        parser_ondemand.add_argument(
            'backup_id',
            type=int,
            metavar='ID',
            help="Id of a backup known to server-side-backups server"
        )

        # get arguments
        cli_args = parser.parse_args()

        # get registration code first because everytime is needed
        registration_code = get_registration_code(cli_args.reg_code_path)

        if cli_args.action == 'create':

            response = backup_create(
                registration_code,
                cli_args.url,
                cli_args.backup,
                cli_args.passwd,
                connection_timeout=cli_args.connection_timeout
            )

        elif cli_args.action == 'retrieve':

            backup_retrieve(
                registration_code,
                cli_args.url,
                cli_args.backup_id,
                cli_args.target,
                cli_args.passwd,
                connection_timeout=cli_args.connection_timeout
            )

            # returned text from this script should be JSON and path to backup is important information
            response = '{"path": "%(path)s"}' % {'path': cli_args.target}

        elif cli_args.action == 'delete':

            call_rest_api(
                registration_code,
                cli_args.url,
                cli_args.action,
                backup_id=cli_args.backup_id,
                connection_timeout=cli_args.connection_timeout
            )
            # returned text from this script should be JSON
            response = '{"deleted": true}'

        elif cli_args.action == 'ondemand':

            response = call_rest_api(
                registration_code,
                cli_args.url,
                cli_args.action,
                backup_id=cli_args.backup_id,
                connection_timeout=cli_args.connection_timeout
            )
        elif cli_args.action == 'list':
            # list
            response = call_rest_api(
                registration_code,
                cli_args.url,
                cli_args.action,
                connection_timeout=cli_args.connection_timeout
            )

        stdout.write(response)

    except SSBackupsException as exc:
        stdout.write(json.dumps({'detail': exc.detail, 'exit_code': exc.code, 'type': exc.type}))
        exit_code = exc.code

    except Exception as exc: # pylint: disable=broad-except
        # handle unhandled exceptions
        stdout.write(json.dumps({'detail': exc.message, 'exit_code': ERR_CODE_UNKNOWN, 'type': exc.__class__.__name__}))
        exit_code = ERR_CODE_UNKNOWN

    stdout.flush()

    exit(exit_code)
